<?php

namespace Drupal\aws_cloud\Service\AwsCloud;

use Drupal\aws_cloud\Entity\Ec2\ElasticIpInterface;
use Drupal\aws_cloud\Entity\Ec2\ImageInterface;
use Drupal\aws_cloud\Entity\Ec2\Instance;
use Drupal\aws_cloud\Entity\Ec2\InstanceInterface;
use Drupal\aws_cloud\Entity\Ec2\KeyPairInterface;
use Drupal\aws_cloud\Entity\Ec2\SecurityGroup;
use Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface;
use Drupal\aws_cloud\Form\Ec2\InstanceEditForm;
use Drupal\aws_cloud\Plugin\Field\FieldType\IpPermission;
use Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface;
use Drupal\aws_cloud\Traits\AwsCloudEntityCheckTrait;
use Drupal\cloud\Entity\CloudConfig;
use Drupal\cloud\Entity\CloudConfigInterface;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Service\CloudServiceInterface;
use Drupal\cloud\Service\EntityLinkRendererInterface;
use Drupal\cloud\Traits\CloudContentEntityTrait;
use Drupal\Core\Entity\EntityDeleteFormTrait;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Field\FieldItemList;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\Url;

/**
 * Interacts with AWS Cloud using AWS Cloud Service API.
 */
class AwsCloudOperationsService implements AwsCloudOperationsServiceInterface {

  use AwsCloudEntityCheckTrait;
  use CloudContentEntityTrait;
  use EntityDeleteFormTrait;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  private $cloudConfigPluginManager;

  /**
   * The Cloud service.
   *
   * @var \Drupal\cloud\Service\CloudServiceInterface
   */
  private $cloudService;

  /**
   * The AWS Cloud EC2 service.
   *
   * @var \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface
   */
  private $ec2Service;

  /**
   * The entity.
   *
   * @var \Drupal\Core\Entity\EntityInterface
   */
  private $entity;

  /**
   * Entity link renderer object.
   *
   * @var \Drupal\cloud\Service\EntityLinkRendererInterface
   */
  private $entityLinkRenderer;

  /**
   * The Entity Type Manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  private $entityTypeManager;

  /**
   * AwsCloudOperationsService constructor.
   *
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\cloud\Service\CloudServiceInterface $cloud_service
   *   The Cloud service.
   * @param \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface $ec2_service
   *   The AWS Cloud EC2 service.
   * @param \Drupal\cloud\Service\EntityLinkRendererInterface $entity_link_renderer
   *   The entity link render service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The Entity Type Manager.
   */
  public function __construct(
    CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
    CloudServiceInterface $cloud_service,
    Ec2ServiceInterface $ec2_service,
    EntityLinkRendererInterface $entity_link_renderer,
    EntityTypeManagerInterface $entity_type_manager
  ) {
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->cloudService = $cloud_service;
    $this->ec2Service = $ec2_service;
    $this->entityLinkRenderer = $entity_link_renderer;
    $this->entityTypeManager = $entity_type_manager;

    $this->messenger = $this->messenger();
  }

  /**
   * {@inheritdoc}
   */
  public function setEc2Service(mixed $ec2_service): void {
    $this->ec2Service = $ec2_service;
  }

  /**
   * Gets the entity of this form.
   *
   * Provided by \Drupal\Core\Entity\EntityForm.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   The entity.
   */
  public function getEntity(): EntityInterface {
    return $this->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function copyFormItemValues(FieldableEntityInterface $entity, array $form): void {
    /** @var \Drupal\Core\Entity\FieldableEntityInterface */
    $original_entity = $this->entityTypeManager
      ->getStorage($entity->getEntityTypeId())
      ->load($entity->id());

    $item_field_names = [];
    foreach ($form ?: [] as $name => $item) {
      if (!is_array($item)) {
        continue;
      }

      if (isset($item['#type'])
        && $item['#type'] === 'item'
        && (!isset($item['#not_field']) || $item['#not_field'] === FALSE)
      ) {
        $item_field_names[] = $name;
      }

      if (isset($item['#type']) && $item['#type'] === 'details') {
        foreach ($item ?: [] as $sub_item_name => $sub_item) {
          if (is_array($sub_item)
            && isset($sub_item['#type'])
            && $sub_item['#type'] === 'item'
            && (!isset($sub_item['#not_field']) || $sub_item['#not_field'] === FALSE)
          ) {
            $item_field_names[] = $sub_item_name;
          }
        }
      }
    }

    foreach ($item_field_names ?: [] as $item_field_name) {
      // Support multi-valued item fields.
      $values = !empty($original_entity) ? $original_entity->get($item_field_name)->getValue() : [];
      if (!empty($values) && count($values) > 1) {
        $item_field_values = array_map(static function ($value) {
          return $value['value'];
        }, $values);
        $entity->set($item_field_name, $item_field_values);
      }

      $item_field_value = !empty($original_entity) && !empty($original_entity->get($item_field_name))
        ? $original_entity->get($item_field_name)->value
        : [];
      if (!empty($item_field_value)) {
        $entity->set($item_field_name, $item_field_value);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function trimTextfields(FieldableEntityInterface $entity, array $form, FormStateInterface $form_state): void {
    $field_names = [];
    foreach ($form ?: [] as $name => $item) {
      if (!is_array($item)) {
        continue;
      }

      if (isset($item['#type'])
        && $item['#type'] === 'textfield'
      ) {
        $field_names[] = $name;
      }

      if (isset($item['#type']) && $item['#type'] === 'details') {
        foreach ($item ?: [] as $sub_item_name => $sub_item) {
          if (is_array($sub_item)
            && isset($sub_item['#type'])
            && $sub_item['#type'] === 'textfield'
          ) {
            $field_names[] = $sub_item_name;
          }
        }
      }
    }

    foreach ($field_names ?: [] as $field_name) {
      $value = $form_state->getValue($field_name);
      if ($value === NULL) {
        continue;
      }

      $value = trim($value);
      $form_state->setValue($field_name, $value);
      $entity->set($field_name, $value);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function saveAwsCloudContent(FieldableEntityInterface $entity, array $form, FormStateInterface $form_state): void {
    // Copy values of the form elements whose type are item to entity.
    // If not, the properties corresponding to the form elements
    // will be saved as NULL.
    $this->copyFormItemValues($entity, $form);

    $this->trimTextfields($entity, $form, $form_state);

    $this->saveCloudContent($entity);
  }

  /**
   * {@inheritdoc}
   */
  public function setTagsInAws(EntityInterface $entity, $resource_id, array $tag_map, $entity_type = ''): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface $entity */
    $this->ec2Service->setCloudContext($entity->getCloudContext());
    $tags = [];
    foreach ($tag_map ?: [] as $key => $value) {
      $tags[] = [
        'Key' => $key,
        'Value' => $value,
      ];
    }

    if (!method_exists($this->ec2Service, 'createTags')) {
      return;
    }

    // Create Tags with different parameters for AWS and OpenStack.
    if (preg_match('[^aws_cloud]', $entity->getEntityTypeId()) === 1) {
      $this->ec2Service->createTags([
        'Resources' => [$resource_id],
        'Tags' => $tags,
      ]);
    }
    else {
      $this->ec2Service->createTags([
        'Resources' => [$resource_id],
        'Tags' => $tags,
        'EntityType' => $entity_type,
      ]);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function updateNameAndCreatedByTags(EntityInterface $entity, $aws_cloud_resource_id): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface $entity */
    $cloud_context = $entity->getCloudContext();
    if ($this->isCloudConfigRemote($entity)) {
      // Get original cloud context.
      if (($pos = strrpos($cloud_context, '_')) !== FALSE) {
        $cloud_context = substr($cloud_context, 0, $pos);
      }
    }

    $uid = !empty($entity->getOwner())
      ? $entity->getOwner()->id()
      : 0;
    /** @var \Drupal\openstack\Service\Ec2\OpenStackService $ec2_service */
    $ec2_service = $this->ec2Service;
    $uid_key_name = str_contains($entity->bundle(), 'aws_cloud')
      ? $entity->getEntityTypeId() . '_' . $this->cloudService->getTagKeyCreatedByUid('aws_cloud', $cloud_context)
      : $ec2_service->getTagKeyCreatedByUid(
        'openstack',
        $cloud_context,
        '',
        $entity->getOwner()->id()
      );
    $this->setTagsInAws($entity, $aws_cloud_resource_id, [
      $uid_key_name => $uid,
      'Name' => $entity->getName(),
    ],
    $entity->getEntityTypeId());
  }

  /**
   * {@inheritdoc}
   */
  public function validateAuthorize(SecurityGroupInterface $group): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface $updated_group */
    $updated_group = $this->entityTypeManager
      ->getStorage($group->getEntityTypeId())
      ->load($group->id());

    if ($group->getIpPermission()->count() !== $updated_group->getIpPermission()->count()) {
      $this->messenger->addError(
        $this->t('Error updating inbound permissions for security group @name', [
          '@name' => $group->label(),
        ])
      );
    }

    if (!empty($group->getVpcId())) {
      if ($group->getOutboundPermission()->count() !== $updated_group->getOutboundPermission()->count()) {
        $this->messenger->addError(
          $this->t('Error updating outbound permissions for security group @name', [
            '@name' => $group->label(),
          ])
        );
      }
    }

    if (empty($this->messenger->messagesByType('error'))) {
      // No errors, success.
      $this->processOperationStatus($group, $this->ec2Service->isWorkerResource() ? 'updated remotely' : 'updated');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function formatIpPermissions(FieldItemList $permissions): array {
    $ip_permissions = [];
    foreach ($permissions ?: [] as $permission) {
      if (empty($permission->getSource())) {
        continue;
      }
      $ip_permissions['IpPermissions'][] = $this->formatIpPermissionForAuthorize($permission);
    }
    return $ip_permissions;
  }

  /**
   * {@inheritdoc}
   */
  public function updateIngressEgressPermissions(SecurityGroupInterface $original_entity, SecurityGroup $entity, SecurityGroup $existing_group, bool $revokeIngress = TRUE, bool $revokeEgress = TRUE): void {
    $inbound_result = $this->updatePermissions(
      $original_entity,
      $entity->getGroupId(),
      $existing_group->getIpPermission(),
      'IpPermissions',
      'getIpPermission',
      'revokeSecurityGroupIngress',
      'authorizeSecurityGroupIngress',
      $revokeIngress
    );

    // Update the outbound permissions.  This only applies to
    // VPC security groups.
    $outbound_result = TRUE;
    if (!empty($entity->getVpcId())) {
      $outbound_result = $this->updatePermissions(
        $original_entity,
        $entity->getGroupId(),
        $existing_group->getOutboundPermission(),
        'IpPermissions',
        'getOutboundPermission',
        'revokeSecurityGroupEgress',
        'authorizeSecurityGroupEgress',
        $revokeEgress
      );
    }

    if (!$inbound_result && !$outbound_result) {
      $this->messenger->addError($this->t('Inbound and Outbound Rules could not be saved because of the error from AWS.'));
    }
    elseif (!$inbound_result) {
      $this->messenger->addError($this->t('Inbound Rules could not be saved because of the error from AWS.'));
    }
    elseif (!$outbound_result) {
      $this->messenger->addError($this->t('Outbound Rules could not be saved because of the error from AWS.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getPermission(string $type, int $position, SecurityGroupInterface $entity) {
    $permission = FALSE;
    switch ($type) {
      case 'outbound_permission':
        $permission = $entity->getOutboundPermission()->get($position);
        break;

      case 'ip_permission':
        $permission = $entity->getIpPermission()->get($position);
        break;

      default:
        break;
    }
    return $permission;
  }

  /**
   * {@inheritdoc}
   */
  public function associateElasticIp(ElasticIpInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      $this->ec2Service->setCloudContext($entity->getCloudContext());

      // Get module name.
      $module_name = $this->getModuleName($entity);

      $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
        'cloud_context' => $entity->getCloudContext(),
        "{$entity->getEntityTypeId()}" => $entity->id(),
      ]);

      // If an IP address is attaching to instance.
      if ($form_state->getValue('resource_type') === 'instance') {

        $instance_entity_id = $form_state->getValue('instance_id');
        $private_ip = $form_state->getValue('instance_private_ip');

        if ($instance_entity_id !== -1) {

          $instances = $this->entityTypeManager
            ->getStorage("{$module_name}_instance")
            ->loadByProperties([
              'id' => $instance_entity_id,
              'cloud_context' => $entity->getCloudContext(),
            ]);

          if (empty($instances)) {
            $this->messenger->addError(
              $this->t('Unable to load the instance %instance_entity_id.', [
                '%instance_entity_id' => $instance_entity_id,
              ]
            ));

            $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
              'cloud_context' => $entity->getCloudContext(),
              "{$entity->getEntityTypeId()}" => $entity->id(),
            ]);
            return FALSE;
          }

          /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $instance */
          $instance = array_shift($instances);
          $instance_id = $instance->getInstanceId();

          $instance_link = $this->entityLinkRenderer->renderViewElement(
            $instance_id,
            "{$module_name}_instance",
            'instance_id',
            [],
            $instance->getName() !== $instance->getInstanceId()
              ? $this->t('@instance_name (@instance_id)', [
                '@instance_name' => $instance->getName(),
                '@instance_id' => $instance_id,
              ])
              : $instance_id
          );

          // Refresh network interfaces.
          $this->ec2Service->updateNetworkInterfaces();

          $network_interface = $this->getNetworkInterfaceByPrivateIp($private_ip, $module_name, $entity->getCloudContext());
          if (!empty($network_interface)) {
            $result = $this->ec2Service->associateAddress([
              'AllocationId' => $entity->getAllocationId(),
              'NetworkInterfaceId' => $network_interface->getNetworkInterfaceId(),
              'PrivateIpAddress' => $private_ip,
            ]);

            if ($result === NULL) {
              $this->messenger->addError($this->t('Unable to associate @label.', [
                '@label' => $entity->getEntityType()->getSingularLabel(),
              ]));
              return FALSE;
            }

            if (!empty($result['SendToWorker'])) {
              $message = $this->t("@label @ip_address associated with @private_ip for instance: @instance_id remotely", [
                '@ip_address' => $entity->getPublicIp(),
                '@private_ip' => $private_ip,
                '@instance_id' => Markup::create($instance_link['#markup']),
                '@label' => $entity->getEntityType()->getSingularLabel(),
              ]);
              $this->messenger->addStatus($message);
              return TRUE;
            }

            $message = $this->t("@label @ip_address associated with @private_ip for instance: @instance_id", [
              '@ip_address' => $entity->getPublicIp(),
              '@private_ip' => $private_ip,
              '@instance_id' => Markup::create($instance_link['#markup']),
              '@label' => $entity->getEntityType()->getSingularLabel(),
            ]);

            $this->updateElasticIpEntity($message);
            $this->clearCacheValues($entity->getCacheTags());
            $this->dispatchSubmitEvent($entity);
            return TRUE;
          }
          else {
            $this->messenger->addError($this->t('Unable to load network interface by private IP.'));
            return FALSE;
          }
        }
        else {
          $this->messenger->addError($this->t('Unable to load instance ID. No association performed.'));
          return FALSE;
        }
      }

      // If an IP address is attaching to network_interface.
      $network_interface_id = $form_state->getValue('network_interface_id');
      $network_private_ip = $form_state->getValue('network_private_ip');

      if ($network_interface_id !== -1) {

        $network_interfaces = $this->entityTypeManager
          ->getStorage("{$module_name}_network_interface")
          ->loadByProperties(
            [
              'id' => $network_interface_id,
              'cloud_context' => $entity->getCloudContext(),
            ]
          );

        /** @var \Drupal\aws_cloud\Entity\Ec2\NetworkInterfaceInterface $network_interface */
        $network_interface = count($network_interfaces) === 1
          ? array_shift($network_interfaces)
          : $network_interfaces;

        $result = $this->ec2Service->associateAddress([
          'AllocationId' => $entity->getAllocationId(),
          'NetworkInterfaceId' => $network_interface->getNetworkInterfaceId(),
          'PrivateIpAddress' => $network_private_ip,
        ]);

        if ($result === NULL) {
          $this->messenger->addError($this->t("Unable to associate @label.", [
            '@label' => $entity->getEntityType()->getSingularLabel(),
          ]));
          return FALSE;
        }

        if (!empty($result['SendToWorker'])) {
          $message = $this->t('@label @ip_address associated with @private_ip for network interface: @network_interface_id remotely', [
            '@ip_address' => $entity->getPublicIp(),
            '@network_interface_id' => $network_interface->getNetworkInterfaceId(),
            '@private_ip' => $network_private_ip,
            '@label' => $entity->getEntityType()->getSingularLabel(),
          ]);
          $this->messenger->addStatus($message);
          return TRUE;
        }

        $message = $this->t('@label @ip_address associated with @private_ip for network interface: @network_interface_id', [
          '@ip_address' => $entity->getPublicIp(),
          '@network_interface_id' => $network_interface->getNetworkInterfaceId(),
          '@private_ip' => $network_private_ip,
          '@label' => $entity->getEntityType()->getSingularLabel(),
        ]);

        $this->updateElasticIpEntity($message);
        $this->clearCacheValues($entity->getCacheTags());
        $this->dispatchSubmitEvent($entity);
        return TRUE;
      }
      else {
        $this->messenger->addError($this->t('Unable to load instance ID. No association performed.'));
        return FALSE;
      }
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'associated');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function associateElasticIpForInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      $this->ec2Service->setCloudContext($entity->getCloudContext());
      $allocation_id = $form_state->getValue('allocation_id');
      $network_interface_id = $form_state->getValue('network_interface_id');

      $result = $this->ec2Service->associateAddress([
        'AllocationId' => $allocation_id,
        'NetworkInterfaceId' => $network_interface_id,
      ]);

      if ($result !== NULL) {
        $this->ec2Service->updateElasticIps();
        $this->ec2Service->updateInstances();
        $this->ec2Service->updateNetworkInterfaces();

        $elastic_ip = $this->getElasticIp($allocation_id, $entity->getCloudContext());
        if ($elastic_ip !== FALSE) {
          $this->messenger->addStatus($this->t('Elastic IP @ip_address associated with @private_ip for instance: @instance', [
            '@ip_address' => $elastic_ip->getPublicIp(),
            '@instance' => $entity->getName(),
            '@private_ip' => $elastic_ip->getPrivateIpAddress(),
          ]));
          $this->clearCacheValues($entity->getCacheTags());
        }
        $form_state->setRedirect('entity.aws_cloud_instance.canonical', [
          'cloud_context' => $entity->getCloudContext(),
          'aws_cloud_instance' => $entity->id(),
        ]);
        return TRUE;
      }
      else {
        $this->messenger->addError($this->t('Unable to associate Elastic IP.'));
        $form_state->setRedirect('entity.aws_cloud_instance.canonical', [
          'cloud_context' => $entity->getCloudContext(),
          'aws_cloud_instance' => $entity->id(),
        ]);
        return FALSE;
      }
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'associated');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createImage(ImageInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      $this->trimTextfields($entity, $form, $form_state);

      $result = $this->ec2Service->createImage([
        'InstanceId'  => $entity->getInstanceId(),
        'Name'        => $entity->getName(),
        'Description' => $entity->getDescription(),
        'Visibility'  => $entity->getVisibility(),
      ]);

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'created remotely');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }

      $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
      $account_id = $this->cloudConfigPluginManager->loadConfigEntity()->get('field_account_id')->value;

      if (!empty($entity) && !empty($result['ImageId'])
        && ($entity->setName($form_state->getValue('name')))
        && ($entity->set('ami_name', $form_state->getValue('name')))
        && ($entity->setImageId($result['ImageId']))
        && ($entity->set('account_id', $account_id))
        && ($entity->save())) {

        $this->updateNameAndCreatedByTags($entity, $entity->getImageId());

        $this->processOperationStatus($entity, 'created');

        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        $this->dispatchSaveEvent($entity);

        return TRUE;
      }
      else {
        $this->processOperationErrorStatus($entity, 'created');

        return FALSE;
      }
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'created');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createImageFromInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }

      $this->ec2Service->setCloudContext($entity->getCloudContext());
      $result = $this->ec2Service->createImage([
        'InstanceId' => $entity->getInstanceId(),
        'Name' => $form_state->getValue('image_name'),
        'NoReboot' => $form_state->getValue('no_reboot') === 0 ? FALSE : TRUE,
      ]);

      if (empty($entity) || empty($result['ImageId'])) {
        // Give an error.
        $this->processOperationErrorStatus($entity, 'created');
        return FALSE;
      }

      // Call image update on this particular image.
      $this->ec2Service->updateImages([
        'ImageIds' => [
          $result['ImageId'],
        ],
      ]);

      $this->messenger()->addStatus($this->t('The @type %label (%image_id) has been created.', [
        '@type' => $entity->getEntityType()->getSingularLabel(),
        '%label' => $entity->toLink($entity->label())->toString(),
        '%image_id' => $result['ImageId'],
      ]));
      $this->clearCacheValues($entity->getCacheTags());
      $image = $this->loadNewImage($entity->getCloudContext(), $result['ImageId']);
      if (!empty($image)) {
        $this->dispatchSubmitEvent($image);
      }
      $form_state->setRedirect('view.aws_cloud_image.list', ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'created');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createKeyPair(KeyPairInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->trimTextfields($entity, $form, $form_state);

      $uid_key_name = $this->cloudService->getTagKeyCreatedByUid(
        'aws_cloud',
        $entity->getCloudContext());

      $result = $this->ec2Service->createKeyPair([
        'KeyName' => $entity->getKeyPairName(),
        'TagSpecifications' => [
          [
            'ResourceType' => 'key-pair',
            'Tags' => [
              [
                'Key' => $entity->getEntityTypeId() . '_' . $uid_key_name,
                'Value' => !empty($entity->getOwner()) ? $entity->getOwner()->id() : 0,
              ],
            ],
          ],
        ],
      ]);

      if (empty($result)) {
        // Use the custom message since key pair uses 'key_pair_name'
        // for its own label.  So do not change the following code.
        $this->messenger->addStatus($this->t('The @type @label could not be created.', [
          '@type' => $entity->getEntityType()->getSingularLabel(),
          '@label' => $entity->getKeyPairName(),
        ]));
        $this->logOperationErrorMessage($entity, 'created');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'created remotely');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }

      $entity->setKeyFingerprint($result['KeyFingerprint']);
      $entity->setKeyPairId($result['KeyPairId']);
      $entity->setKeyMaterial($result['KeyMaterial']);
      $entity->save();
      $this->processOperationStatus($entity, 'created');
      $this->logOperationMessage($entity, 'created');
      $this->clearCacheValues($entity->getCacheTags());

      $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
        'cloud_context' => $entity->getCloudContext(),
        $entity->getEntityTypeId() => $entity->id(),
      ]);
      $this->dispatchSaveEvent($entity);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'created');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createSecurityGroup(SecurityGroupInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      $this->trimTextfields($entity, $form, $form_state);

      $result = $this->ec2Service->createSecurityGroup(
        [
          'GroupName'   => $entity->getGroupName(),
          'Description' => $entity->getDescription(),
        ]
        + ($entity->getEntityTypeId() === 'aws_cloud_security_group'
          ? ['VpcId' => $entity->getVpcId()]
          : []
        )
      );

      if (!empty($result['SendToWorker'])) {
        $entity->setName($entity->getGroupName());
        $this->processOperationStatus($entity, 'created remotely');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }

      if (isset($result['GroupId'])
      && ($entity->setGroupId($result['GroupId']))
      && ($entity->set('name', $entity->getGroupName()))
      && ($entity->save())) {
        $this->updateNameAndCreatedByTags($entity, $entity->getGroupId());

        // Have the system refresh the security group.
        $this->ec2Service->updateSecurityGroups([
          'GroupIds' => [$entity->getGroupId()],
        ], FALSE);

        $this->processOperationStatus($entity, 'created');
        $this->messenger->addStatus('Setup the IP permissions.');

        $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
          'cloud_context' => $entity->getCloudContext(),
          $entity->getEntityTypeId() => $entity->id(),
        ]);
        $this->clearCacheValues($entity->getCacheTags());
        $this->dispatchSaveEvent($entity);
      }
      else {
        // Use the custom message since security group uses 'group_name' for its
        // own label.  So do not change the following code.
        $message = $this->t('The @type @label could not be created.', [
          '@type'  => $entity->getEntityType()->getSingularLabel(),
          '@label' => $entity->getGroupName(),
        ]);
        $this->messenger->addError($message);
        $this->logOperationErrorMessage($entity, 'created');
        return FALSE;
      }

      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'created');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deleteImage(ImageInterface $entity, array $form, FormStateInterface $form_state): bool {
    $this->entity = $entity;
    if ($entity->getEntityTypeId() === 'aws_cloud_image'
        && !$this->updateEntitySubmitForm($form_state, $entity)) {
      return FALSE;
    }

    $this->ec2Service->setCloudContext($entity->getCloudContext());
    $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
    $account_id = $this->cloudConfigPluginManager->loadConfigEntity()->get('field_account_id')->value;

    // If the image is not owned by the aws user,
    // the calling for deregisterImage will be skipped for AWS Cloud,
    // and it will be called for OpenStack.
    if (($entity->getEntityTypeId() === 'aws_cloud_image'
    && ($entity->getAccountId() !== $account_id))
    || ($result = $this->ec2Service->deregisterImage([
      'ImageId' => $entity->getImageId(),
    ])) !== NULL) {

      if (!empty($result['SendToWorker'])) {
        $this->messenger->addStatus($this->t('The @type @label has been deleted remotely.', [
          '@type'  => $entity->getEntityType()->getSingularLabel(),
          '@label' => $entity->getName(),
        ]));

        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }

      $entity->delete();

      // Do not change the following message since we cannot use
      // $entity->label(), which represents 'ami_name'.
      $this->messenger->addStatus($this->t('The @type @label has been deleted.', [
        '@type'  => $entity->getEntityType()->getSingularLabel(),
        '@label' => $entity->getName(),
      ]));
      $this->logDeletionMessage();
      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSubmitEvent($entity);
    }
    else {

      // Do not change the following message since we cannot use
      // $entity->label(), which represents 'ami_name'.
      $this->messenger->addError($this->t('The @type @label could not be deleted.', [
        '@type'  => $entity->getEntityType()->getSingularLabel(),
        '@label' => $entity->getName(),
      ]));

      $this->logger($entity->getEntityType()->getProvider())->error($this->t('@type: @label could not be deleted.', [
        '@type' => $entity->getEntityType()->getLabel(),
        '@label' => $entity->label(),
      ]), [
        'link' => $entity->toLink($this->t('View'))->toString(),
      ]);
    }

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }
      $this->ec2Service->setCloudContext($entity->getCloudContext());

      $result = $this->ec2Service->terminateInstances([
        'InstanceIds' => [$entity->getInstanceId()],
      ]);

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'deleted remotely');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }

      if (!isset($result['TerminatingInstances'][0]['InstanceId'])) {
        $this->processOperationErrorStatus($entity, 'deleted');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return FALSE;
      }

      $entity->delete();

      $this->entity = $entity;
      $this->messenger->addStatus($this->getDeletionMessage());
      $this->logDeletionMessage();
      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSubmitEvent($entity);
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'deleted');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deleteKeyPair(KeyPairInterface $entity, array $form, FormStateInterface $form_state): bool {
    $this->entity = $entity;
    $this->ec2Service->setCloudContext($entity->getCloudContext());

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    $result = $this->ec2Service->deleteKeyPair([
      'KeyName' => $entity->getKeyPairName(),
    ]);

    // If $result is NULL, some error should have occurred
    // when calling the AWS API.
    if ($result === NULL) {
      $this->processOperationErrorStatus($entity, 'deleted');
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $this->messenger->addStatus($this->t('The @type @label has been deleted remotely.', [
        '@type'  => $entity->getEntityType()->getSingularLabel(),
        '@label' => $entity->getName(),
      ]));
      return TRUE;
    }

    $entity->delete();

    $this->messenger->addStatus($this->getDeletionMessage());
    $this->logDeletionMessage();
    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSubmitEvent($entity);
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function editImage(ImageInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface $original_entity */
      $original_entity = $this->entityTypeManager
        ->getStorage($entity->getEntityTypeId())
        ->load($entity->id());

      $this->saveAwsCloudContent($entity, $form, $form_state);

      $this->updateNameAndCreatedByTags($entity, $entity->getImageId());

      $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
      $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();

      // Call 'modifyImageAttribute'
      // only when OwnerId is same as Account ID of Cloud Config.
      // It's to avoid AuthFailure error from AWS API.
      if ($cloud_config->get('field_account_id')->value !== $entity->getAccountId()) {
        return TRUE;
      }

      $form_values = $form_state->getValues();
      $visibility = $form_values['visibility'] ?? 0;

      $permission = [
        [
          'Group' => 'all',
          'UserId' => $entity->getAccountId(),
        ],
      ];

      // 1 => Public, 0 => Private.
      $launch_permission =
        $visibility
          ? ['Add' => $permission]
          : ['Remove' => $permission];

      // Update image.
      $this->ec2Service->setCloudContext($entity->getCloudContext());

      if (!method_exists($this->ec2Service, 'modifyImageAttribute')) {
        return TRUE;
      }

      // Update image description.
      $this->ec2Service->modifyImageAttribute([
        'ImageId' => $entity->getImageId(),
        'Description' => ['Value' => $entity->getDescription()],
      ]);

      // These fields (Description , LaunchPermission , or ProductCode) cannot
      // be specified at the same time.
      // See also:
      // https://docs.aws.amazon.com/cli/latest/reference/ec2/modify-image-attribute.html
      // Update image launch permission.
      $this->ec2Service->modifyImageAttribute([
        'ImageId' => $entity->getImageId(),
        'LaunchPermission' => $launch_permission,
      ]);

      $new_account_ids = [];
      foreach ($entity->getLaunchPermissionAccountIds() ?: [] as $account_id) {
        if (empty($account_id->value)) {
          continue;
        }
        $new_account_ids[] = $account_id->value;
      }

      $old_account_ids = [];
      foreach ($original_entity->getLaunchPermissionAccountIds() ?: [] as $account_id) {
        if (empty($account_id->value)) {
          continue;
        }
        $old_account_ids[] = $account_id->value;
      }

      $account_ids_to_remove = array_diff($old_account_ids, $new_account_ids);
      $account_ids_to_add = array_diff($new_account_ids, $old_account_ids);

      if (!empty($account_ids_to_remove)) {
        $this->ec2Service->modifyImageAttribute([
          'ImageId' => $entity->getImageId(),
          'LaunchPermission' => [
            'Remove' => array_map(function ($item) {
              return ['UserId' => $item];
            }, $account_ids_to_remove),
          ],
        ]);
      }

      if (!empty($account_ids_to_add)) {
        $this->ec2Service->modifyImageAttribute([
          'ImageId' => $entity->getImageId(),
          'LaunchPermission' => [
            'Add' => array_map(function ($item) {
              return ['UserId' => $item];
            }, $account_ids_to_add),
          ],
        ]);
      }

      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSaveEvent($entity);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'edited');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function editInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      $termination_timestamp = !empty($form_state->getValue('termination_timestamp')[0]['value'])
        ? $form_state->getValue('termination_timestamp')[0]['value']->getTimeStamp()
        : NULL;

      if ($termination_timestamp === NULL) {
        // Unset the termination timestamp.
        $entity->set('termination_timestamp', NULL);
      }

      $security_groups = array_values($form_state->getValue('security_groups'));
      if ($security_groups === NULL) {
        $security_groups = [];
      }
      $entity->setSecurityGroups(implode(InstanceEditForm::SECURITY_GROUP_DELIMITER, $security_groups));

      $this->saveAwsCloudContent($entity, $form, $form_state);

      if (!$form_state->getValue('user_data_base64_encoded')) {
        $entity->setUserData(base64_encode($entity->getUserData() ?? ''));
        $entity->save();
      }

      $user_data_default_value = $form['options']['user_data_base64_encoded']['#default_value']
        ? $form['options']['user_data']['#default_value']
        : base64_encode($form['options']['user_data']['#default_value'] ?? '');

      $config = \Drupal::config('aws_cloud.settings');

      $this->updateTagsField(
        $entity,
        $form_state->getValue('tags'),
        $termination_timestamp,
        $config->get('aws_cloud_scheduler') === TRUE
          ? $entity->getSchedule()
          : NULL
      );

      $this->ec2Service->setCloudContext($entity->getCloudContext());

      $this->updateAwsTags($entity);

      // Update the instance type.
      $this->updateInstanceType($entity, NULL, $form['instance']['instance_type']['#default_value']);

      // Update security group.
      $security_groups = array_values($form_state->getValue('security_groups'));
      if ($security_groups === NULL) {
        $security_groups = [];
      }
      $this->updateSecurityGroup(
        $entity,
        InstanceEditForm::SECURITY_GROUP_DELIMITER,
        $form['network']['security_groups']['#default_value'],
        $security_groups
      );

      // Update terminate protection.
      $this->updateTerminateProtection($entity, NULL, $form['options']['termination_protection']['#default_value']);

      // Update IAM role.
      $this->updateIamRole($entity);

      // Update user data.
      $this->updateUserData($entity, NULL, $user_data_default_value);

      // If Elastic IP is specified and instance is stopped, attach it.
      $this->attachElasticIp(
        $entity,
        $form_state->getValue('add_new_elastic_ip'),
        $form_state->getValue('current_allocation_id'),
        $form_state->getValue('current_association_id')
      );

      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSaveEvent($entity);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'edited');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function editKeyPair(KeyPairInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      // Call copyFormItemValues() to ensure the form array is intact.
      $this->copyFormItemValues($entity, $form);

      $this->trimTextfields($entity, $form, $form_state);

      $bundle = str_contains($entity->bundle(), 'aws_cloud')
        ? 'aws_cloud' : 'openstack';

      $uid_key_name = $this->cloudService->getTagKeyCreatedByUid(
        $bundle,
        $entity->getCloudContext());

      $this->ec2Service->setCloudContext($entity->getCloudContext());

      if ($entity->save()) {
        $this->setTagsInAws($entity, $entity->getKeyPairName(), [
          $entity->getEntityTypeId() . '_' . $uid_key_name => !empty($entity->getOwner())
          ? $entity->getOwner()->id() : 0,
        ], $entity->getEntityTypeId());
      }

      $this->processOperationStatus($entity, 'updated');
      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSaveEvent($entity);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'edited');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function editSecurityGroup(SecurityGroupInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      // Call copyFormItemValues() to ensure the form array is intact.
      $this->copyFormItemValues($entity, $form);

      $this->trimTextfields($entity, $form, $form_state);

      $this->ec2Service->setCloudContext($entity->getCloudContext());

      $existing_group = $this->entityTypeManager
        ->getStorage($entity->getEntityTypeId())
        ->load($entity->id());

      if ($entity->save()) {

        $this->updateNameAndCreatedByTags($entity, $entity->getGroupId());

        // Update the inbound/outbound permissions.
        $this->updateIngressEgressPermissions($entity, $entity, $existing_group);

        // Have the system refresh the security group.
        $this->ec2Service->updateSecurityGroups([
          'GroupIds' => [$entity->getGroupId()],
        ], FALSE);

        if (empty($this->messenger->messagesByType('error'))) {
          // Check API calls, see if the permissions updates were
          // successful or not.
          $this->validateAuthorize($entity);
          $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
            'cloud_context' => $entity->getCloudContext(),
            $entity->getEntityTypeId() => $entity->id(),
          ]);
        }
        $this->clearCacheValues($entity->getCacheTags());
        $this->dispatchSaveEvent($entity);
        return TRUE;
      }
      else {
        $this->processOperationErrorStatus($this->entity, 'updated');
        return FALSE;
      }
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'edited');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getConsoleOutput(InstanceInterface $entity): string {
    $this->ec2Service->setCloudContext($entity->getCloudContext());
    $result = $this->ec2Service->getConsoleOutput(['InstanceId' => $entity->getInstanceId()]);

    $output = !empty($result) && !empty($result['Output'])
      ? base64_decode($result['Output'])
      : '';

    return !empty($output) ? $output : '';
  }

  /**
   * {@inheritdoc}
   */
  public function importKeyPair(KeyPairInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      $uid_key_name = $this->cloudService->getTagKeyCreatedByUid(
        'aws_cloud',
        $entity->getCloudContext());

      if (!($path = $form_state->getValue('key_pair_public_key'))) {
        return FALSE;
      }

      $handle = fopen($path, 'r');
      $key_material = fread($handle, filesize($path));
      fclose($handle);

      $result = $this->ec2Service->importKeyPair([
        'KeyName' => $entity->getKeyPairName(),
        'PublicKeyMaterial' => $key_material,
        'TagSpecifications' => [
          [
            'ResourceType' => 'key-pair',
            'Tags' => [
              [
                'Key' => $entity->getEntityTypeId() . '_' . $uid_key_name,
                'Value' => !empty($entity->getOwner()) ? $entity->getOwner()->id() : 0,
              ],
            ],
          ],
        ],
      ]);

      if (empty($result)) {
        $this->processOperationErrorStatus($entity, 'imported');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'imported remotely');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }

      // Following AWS specification and not storing key material.
      if ((!empty($entity)
      && $entity->setKeyFingerprint($result['KeyFingerprint'])
      && $entity->setKeyPairId($result['KeyPairId'])
      && $entity->save())) {

        // Use the custom message since key pair uses 'key_pair_name' for its
        // own label.  So do not change the following code.
        $this->messenger->addStatus($this->t('The @type %label has been imported.', [
          '@type' => $entity->getEntityType()->getSingularLabel(),
          '%label' => $entity->toLink($entity->getKeyPairName())->toString(),
        ]));
        $this->logOperationMessage($entity, 'imported');

        $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
          'cloud_context' => $entity->getCloudContext(),
          $entity->getEntityTypeId() => $entity->id(),
        ]);

        $this->clearCacheValues($entity->getCacheTags());
        $this->dispatchSaveEvent($entity);
        return TRUE;
      }

      $this->processOperationErrorStatus($entity, 'imported');

      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return FALSE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'imported');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function launchInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      // Set parameters to entity.
      $params = [
        'image_id'          => $form['image_id']['#value'],
        'max_count'         => $form['max_count']['#value'],
        'min_count'         => $form['min_count']['#value'],
        'key_pair_name'     => $form['key_pair_name']['#value'],
        'is_monitoring'     => $form['is_monitoring']['#value'] ? TRUE : FALSE,
        'availability_zone' => $form['availability_zone']['#value'],
        'security_groups'   => $form['security_groups']['#value'],
        'instance_type'     => $form['instance_type']['#value'],
        'kernel_id'         => $form['kernel_id']['#value'],
        'ramdisk_id'        => $form['ramdisk_id']['#value'],
        'user_data'         => $form['user_data']['#value'],
      ];

      foreach ($params ?: [] as $key => $value) {
        $entity->set($key, $value);
      }

      // Get parameters for launching an instance.
      $key_name       = preg_replace('/ \([^)]*\)$/', '', $entity->getKeyPairName());
      $security_group = preg_replace('/ \([^)]*\)$/', '', $entity->getSecurityGroups());

      $params = [
        // The following parameters are required.
        'ImageId'        => $entity->getImageId(),
        'MaxCount'       => $entity->getMaxCount(),
        'MinCount'       => $entity->getMinCount(),
        'InstanceType'   => $entity->getInstanceType(),
        'Monitoring'     => ['Enabled' => $entity->isMonitoring() ? TRUE : FALSE],
        'KeyName'        => $key_name,
        'Placement'      => ['AvailabilityZone' => $entity->getAvailabilityZone()],
        'SecurityGroups' => [$security_group],
      ];

      // The following parameters are optional.
      $params['KernelId'] ?: $entity->getKernelId();
      $params['RamdiskId'] ?: $entity->getRamdiskId();
      $params['UserData'] ?: $entity->getUserData();

      // Launch an instance.
      $result = $this->ec2Service->runInstances($params, []);

      if (!isset($result['Instances'][0]['InstanceId'])) {
        $this->processOperationErrorStatus($entity, 'launched');
        return FALSE;
      }

      if (($entity->setInstanceId($result['Instances'][0]['InstanceId']))
        && ($entity->setPublicIp($result['Instances'][0]['PublicIpAddress']))
        && ($entity->setKeyPairName($result['Instances'][0]['KeyName']))
        && ($entity->setInstanceState($result['Instances'][0]['State']['Name']))
        && ($entity->setCreated($result['Instances'][0]['LaunchTime']))
        && ($entity->save())) {

        $instance_ids = array_map(static function ($result_instance) {
          return $result_instance['InstanceId'];
        }, $result['Instances']);

        $this->messenger->addStatus($this->t('The @type "@label (@instance_id)" request has been initiated. This may take some time. Use Refresh to update the status.', [
          '@type'        => $entity->getEntityType()->getSingularLabel(),
          '@label'       => $entity->label(),
          '@instance_id' => implode(', ', $instance_ids),
        ]));

        $form_state->setRedirectUrl(
          $entity
            ->toUrl('collection')
            ->setRouteParameter('cloud_context', $entity->getCloudContext())
        );
      }
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'launched');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function rebootInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }

      $this->ec2Service->setCloudContext($entity->getCloudContext());
      $result = $this->ec2Service->rebootInstances([
        'InstanceIds' => [
          $entity->getInstanceId(),
        ],
      ]);

      if (empty($result)) {
        $this->processOperationErrorStatus($entity, 'rebooted');
        $form_state->setRedirect("view.{$this->entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'rebooted remotely');
        $form_state->setRedirect("view.{$this->entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }

      $this->processOperationStatus($entity, 'rebooted');
      $this->clearCacheValues($entity->getCacheTags());
      $form_state->setRedirect("view.{$this->entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'rebooted');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function revokeSecurityGroup(SecurityGroupInterface $entity, array $form, FormStateInterface $form_state, string $type, string $position): bool {
    $this->ec2Service->setCloudContext($entity->getCloudContext());

    $permission = $this->getPermission($type, $position, $entity);
    if (!$permission) {
      return FALSE;
    }

    $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
      'cloud_context' => $entity->getCloudContext(),
      $entity->getEntityTypeId() => $entity->id(),
    ]);

    $perm_array = [
      'GroupId' => $entity->getGroupId(),
    ];

    if ($type === 'outbound_permission') {
      $perm_array['IpPermissions'][] = $this->formatIpPermissionForRevoke($permission);
      $result = $this->ec2Service->revokeSecurityGroupEgress($perm_array);
    }
    else {
      $perm_array['IpPermissions'][] = $this->formatIpPermissionForRevoke($permission);
      $result = $this->ec2Service->revokeSecurityGroupIngress($perm_array);
    }

    if (!empty($result['SendToWorker'])) {
      $this->messenger->addStatus($this->t('Permission revoked remotely'));
      return TRUE;
    }

    // Have the system refresh the security group.
    $this->ec2Service->updateSecurityGroups([
      'GroupIds' => [$entity->getGroupId()],
    ], FALSE);

    $this->validateRevoke($type, $entity)
      ? $this->messenger->addStatus($this->t('Permission revoked'))
      : $this->messenger->addError($this->t('Permission not revoked'));

    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSubmitEvent($entity);

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function startInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }

      $this->ec2Service->setCloudContext($entity->getCloudContext());
      $result = $this->ec2Service->startInstances([
        'InstanceIds' => [
          $entity->getInstanceId(),
        ],
      ]);

      if (empty($result)) {
        $this->processOperationErrorStatus($entity, 'started');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'started remotely');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'started');
      return FALSE;
    }

    try {
      $current_state = $result['StartingInstances'][0]['CurrentState']['Name'];
      /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
      $instance = $this->entityTypeManager
        ->getStorage($entity->getEntityTypeId())
        ->load($entity->id());
      $instance->setInstanceState($current_state);
      $instance->save();

      $this->processOperationStatus($instance, 'started');
      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSubmitEvent($entity);
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'started');
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function stopInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }

      $this->ec2Service->setCloudContext($entity->getCloudContext());
      $result = $this->ec2Service->stopInstances([
        'InstanceIds' => [
          $entity->getInstanceId(),
        ],
      ]);

      if (empty($result)) {
        $this->processOperationErrorStatus($entity, 'stopped');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'stopped remotely');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'stopped');
      return FALSE;
    }

    try {
      $current_state = $result['StoppingInstances'][0]['CurrentState']['Name'];
      /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
      $instance = $this->entityTypeManager
        ->getStorage($entity->getEntityTypeId())
        ->load($entity->id());
      $instance->setInstanceState($current_state);
      $instance->save();

      $this->processOperationStatus($instance, 'stopped');
      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSubmitEvent($entity);
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'stopped');
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return FALSE;
    }
  }

  /**
   * Save a cloud content.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   Interface for entities having fields.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function saveCloudContent(FieldableEntityInterface $entity): int {
    if (!($entity instanceof CloudConfigInterface)) {
      /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $original_entity */
      $original_entity = $entity;
      $this->cloudConfigPluginManager->setCloudContext($original_entity->getCloudContext());
      $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();
    }
    else {
      $cloud_config = CloudConfig::create([
        'cloud_context' => $entity->getCloudContext(),
      ]);
    }

    if (!$entity->isNew() && !($entity instanceof CloudConfigInterface) && $cloud_config->isRemote()) {
      $status = SAVED_UPDATED;
      $passive_operation = 'updated remotely';
      $this->processOperationStatus($entity, $passive_operation);
      return $status;
    }

    $status = $entity->save();

    // Add an updated status message and the log.
    switch ($status) {
      case SAVED_NEW:
        $passive_operation = 'created';
        break;

      case SAVED_DELETED:
        $passive_operation = 'deleted';
        break;

      default:
      case SAVED_UPDATED:
        $passive_operation = 'updated';
    }

    if (!($entity instanceof CloudConfigInterface) && $cloud_config->isRemote()) {
      $passive_operation .= ' remotely';
    }

    $this->processOperationStatus($entity, $passive_operation);

    // Clear block and menu cache.
    $this->clearCacheValues($entity->getCacheTags());

    return $status;
  }

  /**
   * Helper function to update field tags.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   * @param array $form_tags
   *   Form tags.
   * @param int|null $termination_timestamp
   *   Termination timestamp by UNIX seconds.
   * @param string|null $schedule
   *   Schedule.
   */
  private function updateTagsField(
    InstanceInterface $entity,
    array $form_tags,
    ?int $termination_timestamp,
    ?string $schedule
  ): void {
    usort($form_tags, static function ($a, $b) {
      return (int) $a['_weight'] - (int) $b['_weight'];
    });

    $module_name = $this->getModuleName($entity);
    $uid_key_name = $this->cloudService->getTagKeyCreatedByUid(
      'aws_cloud',
      $entity->getCloudContext(),
      Instance::TAG_LAUNCHED_BY_UID
    );

    $fixed_tags = [];
    $fixed_tags['Name'] = $entity->getName();
    $fixed_tags[$module_name . '_' . $uid_key_name] = is_null($entity->getOwner())
      ? 0
      : $entity->getOwner()->id();

    $fixed_tags[$module_name . '_' . InstanceInterface::TAG_TERMINATION_TIMESTAMP] = is_null($termination_timestamp)
      ? ''
      : $termination_timestamp;

    $config = \Drupal::config('aws_cloud.settings');
    if ($config->get('aws_cloud_scheduler') === TRUE) {
      $fixed_tags['aws_cloud_scheduler_tag'] = $schedule;
    }

    $tags = [];
    foreach ($form_tags ?: [] as $form_tag) {
      $item_key = $form_tag['item_key'];
      if ($item_key === '') {
        continue;
      }

      // Skip special tags.
      if (strpos($item_key, 'aws:') === 0) {
        continue;
      }

      $item_value = $form_tag['item_value'];
      if (isset($fixed_tags[$item_key])) {
        $item_value = $fixed_tags[$item_key];
        unset($fixed_tags[$item_key]);
      }

      $tags[] = ['item_key' => $item_key, 'item_value' => $item_value];
    }

    foreach ($fixed_tags ?: [] as $item_key => $item_value) {
      $tags[] = ['item_key' => $item_key, 'item_value' => $item_value];
    }

    $entity->setTags($tags);
    $entity->save();
  }

  /**
   * Helper function to update AWS tags.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   */
  private function updateAwsTags(InstanceInterface $entity): void {
    $this->ec2Service->setCloudContext($entity->getCloudContext());

    $params = [
      'Resources' => [$entity->getInstanceId()],
    ];

    // Delete old tags.
    $this->ec2Service->deleteTags($params);

    foreach ($entity->getTags() ?: [] as $tag) {
      $params['Tags'][] = [
        'Key' => $tag['item_key'],
        'Value' => $tag['item_value'],
      ];
    }

    // Create Tags with different parameters for AWS and OpenStack.
    if (preg_match('[^aws_cloud]', $entity->getEntityTypeId()) === 1) {
      $this->ec2Service->createTags($params);
    }
    else {
      $this->ec2Service->createTags([
        'Resources' => [$entity->getInstanceId()],
        'Tags' => [
          ['Key' => 'Name', 'Value' => $entity->getName()],
        ],
      ]);
    }
  }

  /**
   * Update the instance type.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   * @param string|null $instance_type
   *   Instance type.
   * @param string|null $default_instance_type
   *   Default value of instance type.
   */
  private function updateInstanceType(
    InstanceInterface $entity,
    ?string $instance_type,
    ?string $default_instance_type
  ): void {
    // If the state of the instance is not stopped, do not update it.
    if ($entity->getInstanceState() !== 'stopped') {
      return;
    }

    // When executed from Form, $instance_type is NULL.
    // When executed from REST API, $instance_type is not NULL.
    // If it is run from Form and the instance type of $entity
    // is equal to the default instance type, the instance type
    // is not changed and no update is performed.
    if ($instance_type === NULL
      && $entity->getInstanceType() === $default_instance_type) {
      return;
    }

    $this->ec2Service->modifyInstanceAttribute([
      'InstanceId' => $entity->getInstanceId(),
      'InstanceType' => ['Value' => $instance_type ?? $entity->getInstanceType()],
    ]);
  }

  /**
   * Get security group IDs by Names.
   *
   * @param array $group_names
   *   Array of group names.
   * @param string $vpc_id
   *   VPC id used for the query.
   *
   * @return array
   *   Array of group names.
   */
  private function getSecurityGroupIdsByNames(array $group_names, string $vpc_id): array {
    // Querying for Group Name is done using filters.
    // @see https://docs.aws.amazon.com/aws-sdk-php/v3/api/api-ec2-2016-11-15.html#describesecuritygroups
    $response = $this->ec2Service->describeSecurityGroups([
      'Filters' => [
        [
          'Name' => 'group-name',
          'Values' => $group_names,
        ],
        [
          // Specify VPC ID because the group names can duplicate across
          // different VPC networks. For example the 'default' group name
          // exists across different groups.
          'Name' => 'vpc-id',
          'Values' => [
            $vpc_id,
          ],
        ],
      ],
    ]);

    $group_ids = [];
    foreach ($response['SecurityGroups'] ?: [] as $security_group) {
      $group_ids[] = $security_group['GroupId'];
    }

    return $group_ids;
  }

  /**
   * Update security group.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   * @param string $security_group_delimiter
   *   Delimiter string for security groups.
   * @param array $default_security_groups
   *   Default value of security groups.
   * @param array $security_groups
   *   Security groups.
   */
  private function updateSecurityGroup(
    InstanceInterface $entity,
    string $security_group_delimiter,
    array $default_security_groups,
    array $security_groups
  ): void {
    if ($entity->getSecurityGroups() !== implode($security_group_delimiter, $default_security_groups)) {

      $this->ec2Service->modifyInstanceAttribute([
        'InstanceId' => $entity->getInstanceId(),
        'Groups' => $this->getSecurityGroupIdsByNames($security_groups, $entity->getVpcId()),
      ]);
    }
  }

  /**
   * Update terminate protection.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   * @param bool|null $termination_protection
   *   Termination protection flag.
   * @param bool|null $default_termination_protection
   *   Default value of termination protection flag.
   */
  private function updateTerminateProtection(
    InstanceInterface $entity,
    ?bool $termination_protection,
    ?bool $default_termination_protection
  ): void {
    if ($termination_protection === NULL
      && $entity->getTerminationProtection() === $default_termination_protection) {
      return;
    }

    $this->ec2Service->modifyInstanceAttribute([
      'InstanceId' => $entity->getInstanceId(),
      'DisableApiTermination' => [
        'Value' => $termination_protection ?? $entity->getTerminationProtection(),
      ],
    ]);
  }

  /**
   * Update IAM role.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   */
  private function updateIamRole(InstanceInterface $entity): void {
    if ($entity->getEntityTypeId() === 'aws_cloud_instance') {
      // Update IAM role.
      $associations_result = $this->ec2Service->describeIamInstanceProfileAssociations([
        'Filters' => [
          [
            'Name' => 'instance-id',
            'Values' => [$entity->getInstanceId()],
          ],
        ],
      ]);

      $associations = $associations_result['IamInstanceProfileAssociations'];
      if (empty($associations)) {
        if ($entity->getIamRole() !== NULL) {
          // Associate.
          $this->ec2Service->associateIamInstanceProfile([
            'InstanceId' => $entity->getInstanceId(),
            'IamInstanceProfile' => [
              'Arn' => $entity->getIamRole(),
            ],
          ]);
        }
      }
      else {
        if ($entity->getIamRole() === NULL) {
          // Disassociate.
          $this->ec2Service->disassociateIamInstanceProfile([
            'AssociationId' => $associations[0]['AssociationId'],
          ]);
        }
        elseif ($entity->getIamRole() !== $associations[0]['IamInstanceProfile']['Arn']) {
          // Disassociate.
          $this->ec2Service->disassociateIamInstanceProfile([
            'AssociationId' => $associations[0]['AssociationId'],
          ]);

          // Associate.
          $this->ec2Service->associateIamInstanceProfile([
            'InstanceId' => $entity->getInstanceId(),
            'IamInstanceProfile' => [
              'Arn' => $entity->getIamRole(),
            ],
          ]);
        }
      }
    }
  }

  /**
   * Update user data.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   * @param string|null $user_data
   *   User data.
   * @param string|null $default_user_data
   *   Default value of user data.
   */
  private function updateUserData(InstanceInterface $entity, ?string $user_data, ?string $default_user_data): void {
    if ($entity->getInstanceState() !== 'stopped') {
      return;
    }

    if ($user_data === NULL
      && $entity->getUserData() === $default_user_data) {
      return;
    }

    $user_data_impl = $user_data ?? base64_decode($entity->getUserData() ?? '');
    $this->ec2Service->modifyInstanceAttribute([
      'InstanceId' => $entity->getInstanceId(),
      'UserData' => ['Value' => $user_data_impl],
    ]);

    $url = Url::fromRoute(
      'entity.aws_cloud_instance.start_form', [
        'cloud_context' => $entity->getCloudContext(),
        'aws_cloud_instance' => $entity->id(),
      ]
    )->toString();

    $this->messenger->addStatus($this->t("The user data of @type %label has been updated. <strong><em><a href='@start_instance_link'>Start</a></em></strong> the @type to reflect the user data.", [
      '@type' => $entity->getEntityType()->getSingularLabel(),
      '%label' => $entity->toLink($entity->label())->toString(),
      '@start_instance_link' => $url,
    ]));
    $this->logOperationMessage($entity, 'updated');
  }

  /**
   * If Elastic IP is specified and instance is stopped, attach it.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   * @param string|null $new_elastic_ip
   *   New Elastic IP.
   * @param string|null $current_allocation_id
   *   Current Allocation ID.
   * @param string|null $current_association_id
   *   Current Association ID.
   */
  private function attachElasticIp(
    InstanceInterface $entity,
    ?string $new_elastic_ip,
    ?string $current_allocation_id,
    ?string $current_association_id
  ): void {
    if ($entity->getInstanceState() === 'stopped') {
      if (isset($new_elastic_ip) && (int) $new_elastic_ip !== -1) {
        $update_entities = FALSE;

        if (isset($current_allocation_id) && isset($current_association_id)) {
          // Reassign the Elastic IP address to the instance.
          // (the Elastic IP address = ['AssociationId', 'InstanceId'])
          if ($current_allocation_id !== $new_elastic_ip) {
            $this->ec2Service->disassociateAddress([
              'AssociationId' => $current_association_id,
              'InstanceId' => $entity->getInstanceId(),
            ]);

            $this->ec2Service->associateAddress([
              'AllocationId' => $new_elastic_ip,
              'InstanceId' => $entity->getInstanceId(),
            ]);
            $update_entities = TRUE;
          }
        }
        else {
          // Instance does not have Elastic IP. Assign a new one.
          $this->ec2Service->associateAddress([
            'AllocationId' => $new_elastic_ip,
            'InstanceId' => $entity->getInstanceId(),
          ]);
          $update_entities = TRUE;
        }
        if ($update_entities === TRUE) {
          // Update the following entities if the Elastic IPs have changed.
          $this->ec2Service->updateElasticIps();
          $this->ec2Service->updateInstances();
          $this->ec2Service->updateNetworkInterfaces();
        }
      }
    }
  }

  /**
   * Helper function that loads an aws_cloud_elastic_ip entity.
   *
   * @param string $allocation_id
   *   The allocation ID to look up.
   * @param string $cloud_context
   *   The cloud context.
   *
   * @return \Drupal\aws_cloud\Entity\Ec2\ElasticIpInterface|false
   *   The loaded aws_cloud_elastic_ip entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getElasticIp(string $allocation_id, string $cloud_context): ElasticIpInterface|FALSE {
    $elastic_ip = FALSE;
    $results = $this->entityTypeManager
      ->getStorage('aws_cloud_elastic_ip')
      ->loadByProperties([
        'allocation_id' => $allocation_id,
        'cloud_context' => $cloud_context,
      ]);
    if (count($results) === 1) {
      $elastic_ip = array_shift($results);
    }
    return $elastic_ip;
  }

  /**
   * Helper method to load image entity.
   *
   * @param string $cloud_context
   *   Cloud context.
   * @param string $image_id
   *   Image id to load.
   *
   * @return null|\Drupal\aws_cloud\entity\ec2\ImageInterface
   *   NULL if not found else Image entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function loadNewImage(string $cloud_context, string $image_id): ?ImageInterface {
    $entities = $this->entityTypeManager
      ->getStorage('aws_cloud_image')
      ->loadByProperties([
        'cloud_context' => $cloud_context,
        'image_id' => $image_id,
      ]);

    return !empty($entities)
      ? array_shift($entities)
      : NULL;
  }

  /**
   * Check if current cloud config is remote or not.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Interface for entities having fields.
   *
   * @return bool
   *   True if the current cloud config is remote.
   */
  private function isCloudConfigRemote(EntityInterface $entity): bool {
    /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface $entity */
    $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
    $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();
    return $cloud_config->isRemote();
  }

  /**
   * Format the IpPermission object.
   *
   * Format the IpPermission object for use with
   * the AuthorizeSecurityGroup [Ingress and Egress] Amazon EC2 API call.
   *
   * @param \Drupal\aws_cloud\Plugin\Field\FieldType\IpPermission $ip_permission
   *   The IP permission object.
   *
   * @return array
   *   The permission.
   */
  private function formatIpPermissionForAuthorize(IpPermission $ip_permission): array {
    $permission = [
      'FromPort' => (int) $ip_permission->getFromPort(),
      'ToPort' => (int) $ip_permission->getToPort(),
      'IpProtocol' => $ip_permission->getIpProtocol(),
    ];
    if ($ip_permission->getSource() === 'ip4') {
      $item = [
        'CidrIp' => $ip_permission->getCidrIp(),
      ];
      if (!empty($ip_permission->getDescription())) {
        $item['Description'] = $ip_permission->getDescription();
      }
      $permission['IpRanges'][] = $item;
    }
    elseif ($ip_permission->getSource() === 'ip6') {
      $item = [
        'CidrIpv6' => $ip_permission->getCidrIpv6(),
      ];
      if (!empty($ip_permission->getDescription())) {
        $item['Description'] = $ip_permission->getDescription();
      }
      $permission['Ipv6Ranges'][] = $item;
    }
    elseif ($ip_permission->getSource() === 'prefix') {
      $item = [
        'PrefixListId' => $ip_permission->getPrefixListId(),
      ];
      if (!empty($ip_permission->getDescription())) {
        $item['Description'] = $ip_permission->getDescription();
      }
      $permission['PrefixListIds'][] = $item;
    }
    else {
      $group = [];
      // Use GroupID if non-default VPC or EC2-Classic.
      // For other permissions, use Group Name.
      if (!empty($ip_permission->getGroupId())) {
        $group['GroupId'] = $ip_permission->getGroupId();
      }
      else {
        $group['GroupName'] = $ip_permission->getGroupName();
      }

      $group['UserId'] = $ip_permission->getUserId();
      $group['PeeringStatus'] = $ip_permission->getPeeringStatus();
      $group['VpcId'] = $ip_permission->getVpcId();
      $group['VpcPeeringConnectionId'] = $ip_permission->getPeeringConnectionId();
      if (!empty($ip_permission->getDescription())) {
        $group['Description'] = $ip_permission->getDescription();
      }

      $permission['UserIdGroupPairs'][] = $group;
    }
    return $permission;
  }

  /**
   * Format IP Permissions.
   *
   * @param \Drupal\Core\Field\FieldItemList $ip_permissions
   *   The security group.
   *
   * @return array
   *   Formatted IP Permissions.
   */
  private function formatIpPermissionsForRevoke(FieldItemList $ip_permissions): array {
    $permissions_formatted = [];
    foreach ($ip_permissions ?: [] as $ip_permission) {
      if (empty($ip_permission->getSource())) {
        continue;
      }
      $permission_formatted = $this->formatIpPermissionForAuthorize($ip_permission);
      $permissions_formatted[] = $permission_formatted;
    }

    return $permissions_formatted;
  }

  /**
   * Revoke a set of permissions.
   *
   * @param string $method
   *   String value of revokeSecurityGroupEgress or revokeSecurityGroupIngress.
   * @param array $revoke_permissions
   *   Array of permissions to revoke.
   * @param string $group_id
   *   Security group id to revoke.
   */
  private function revokeIpPermissions($method, array $revoke_permissions, $group_id): void {
    if (!empty($revoke_permissions)) {
      // Delete the existing egress permissions.
      $this->ec2Service->$method([
        'GroupId' => $group_id,
        'IpPermissions' => $revoke_permissions,
      ]);
    }
  }

  /**
   * Helper method to format a permissions array for resubmission.
   *
   * This is used in case of any error during the permissions update process.
   *
   * @param \Drupal\Core\Field\FieldItemList $permissions
   *   IP Permissions.
   * @param string $group_id
   *   Group ID.
   *
   * @return array
   *   Formatted permissions array.
   */
  private function formatIpPermissionsForAdd(FieldItemList $permissions, $group_id): array {
    $existing_permissions['GroupId'] = $group_id;
    foreach ($permissions ?: [] as $permission) {
      if (empty($permission->getSource())) {
        continue;
      }
      $existing_permissions['IpPermissions'][] = $this->formatIpPermissionForAuthorize($permission);
    }
    return $existing_permissions;
  }

  /**
   * Main method for updating permissions.
   *
   * Supports Ingress and Egress permissions.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface $entity
   *   The AWS Cloud instance entity.
   * @param string $group_id
   *   Security group ID.
   * @param \Drupal\Core\Field\FieldItemList $existing_permissions
   *   Existing permissions.
   * @param string $permission_name
   *   The permission field name.
   * @param string $entity_func
   *   The entity function name to retrieve permissions.  Options are
   *   getIpPermission or getOutboundPermission.
   * @param string $revoke_func
   *   The API function name to revoke permissions.  Options are
   *   revokeSecurityGroupIngress or revokeSecurityGroupEgress.
   * @param string $authorize_func
   *   The API function name to authorize permissions.  Options are
   *   authorizeSecurityGroupIngress or authorizeSecurityGroupEgress.
   * @param bool $revoke
   *   Indicates skipping revoke operation.
   *
   * @return bool
   *   Indicates success or failure
   */
  private function updatePermissions(
    SecurityGroupInterface $entity,
    string $group_id,
    FieldItemList $existing_permissions,
    string $permission_name,
    string $entity_func,
    string $revoke_func,
    string $authorize_func,
    bool $revoke = TRUE
  ) : bool {
    // Clear out the existing permissions.
    if (!empty($existing_permissions) && $revoke === TRUE) {
      $revoke_permissions = $this->formatIpPermissionsForRevoke($existing_permissions);
      $this->revokeIpPermissions($revoke_func, $revoke_permissions, $group_id);
    }

    // Format the user entered permissions into AWS accepted array.
    $permissions = $this->formatIpPermissions($entity->$entity_func());
    if (!empty($permissions[$permission_name])) {
      // Setup permissions array for AuthorizeSecurityGroupIngress.
      $permissions['GroupId'] = $group_id;
      $result = $this->ec2Service->$authorize_func($permissions);

      // $result === NULL means an exception was thrown.  Attempt to add
      // the old set of permissions so the user's configurations are not lost.
      if ($result === NULL || count($this->messenger->messagesByType('error')) >= 1) {
        if (!empty($existing_permissions)) {
          $existing_permissions = $this->formatIpPermissionsForAdd(
            $existing_permissions,
            $entity->getGroupId()
          );
          $result = $this->ec2Service->$authorize_func($existing_permissions);
          if ($result === NULL) {
            $this->messenger->addError($this->t('Cannot re-add existing @name rules', [
              '@name' => $permission_name,
            ]));
          }
        }
        return FALSE;
      }
    }
    return TRUE;
  }

  /**
   * Format the IP permission array for use with revoke security group.
   *
   * @param \Drupal\aws_cloud\Plugin\Field\FieldType\IpPermission $ip_permission
   *   IP Permission object.
   *
   * @return array
   *   Permissions array.
   */
  private function formatIpPermissionForRevoke(IpPermission $ip_permission): array {
    $permission = [
      'FromPort' => (int) $ip_permission->getFromPort(),
      'ToPort' => (int) $ip_permission->getToPort(),
      'IpProtocol' => $ip_permission->getIpProtocol(),
    ];
    if ($ip_permission->getSource() === 'ip4') {
      $permission['IpRanges'][]['CidrIp'] = $ip_permission->getCidrIp();
    }
    else {
      if ($ip_permission->getSource() === 'ip6') {
        $permission['Ipv6Ranges'][]['CidrIpv6'] = $ip_permission->getCidrIpv6();
      }
      else {
        // Use GroupID if non-default VPC or EC2-Classic.
        // For other permissions, use Group Name.
        /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface $entity */
        $entity = $this->entity;
        $vpc_id = $entity->getVpcId();
        if ($entity->isDefaultVpc() === TRUE || !isset($vpc_id)) {
          $group['GroupName'] = $ip_permission->getGroupName();
        }
        else {
          $group['GroupId'] = $ip_permission->getGroupId();
        }
        $group['UserId'] = $ip_permission->getUserId();
        $group['PeeringStatus'] = $ip_permission->getPeeringStatus();
        $group['VpcId'] = $ip_permission->getVpcId();
        $group['VpcPeeringConnectionId'] = $ip_permission->getPeeringConnectionId();

        $permission['UserIdGroupPairs'][] = $group;
      }
    }
    return $permission;
  }

  /**
   * Verify that the revoke was successful.
   *
   * Since EC2 does not return any error codes in the revoke API calls, the
   * only way to verify is to count the permissions array from the current
   * entity, and the entity that is newly updated from the updateSecurityGroups
   * API call.
   *
   * @param string $type
   *   The permission type.
   * @param \Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface $group
   *   The security group.
   *
   * @return bool
   *   True or false.
   */
  private function validateRevoke($type, SecurityGroupInterface $group): bool {
    $verified = FALSE;
    /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroup $updated_group */
    $updated_group = $this->entityTypeManager
      ->getStorage($group->getEntityTypeId())
      ->load($group->id());
    if ($type === 'outbound_permission') {
      if ($group->getIpPermission()->count() === $updated_group->getIpPermission()->count()) {
        $verified = TRUE;
      }
    }
    else {
      if ($group->getOutboundPermission()->count() === $updated_group->getOutboundPermission()->count()) {
        $verified = TRUE;
      }
    }
    return $verified;
  }

  /**
   * Get a network interface given a private_ip address.
   *
   * @param string $private_ip
   *   The private IP used to look up the network interface.
   * @param string $module_name
   *   Module name.
   * @param string $cloud_context
   *   Cloud context.
   *
   * @return \Drupal\aws_cloud\Entity\Ec2\NetworkInterfaceInterface|null
   *   False or network interface object.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getNetworkInterfaceByPrivateIp($private_ip, $module_name, $cloud_context) {
    $results = $this->entityTypeManager
      ->getStorage("{$module_name}_network_interface")
      ->loadByProperties([
        'primary_private_ip' => $private_ip,
        'cloud_context' => $cloud_context,
      ]);

    return count($results) === 1
      ? array_shift($results)
      : NULL;
  }

  /**
   * Helper function to update the current aws_cloud_elastic_ip entity.
   *
   * @param string $message
   *   Message to display to use.
   */
  private function updateElasticIpEntity($message): void {
    $this->messenger->addStatus($message);

    // Update the following entities from EC2.
    $this->ec2Service->updateElasticIps();
    $this->ec2Service->updateInstances();
    $this->ec2Service->updateNetworkInterfaces();
  }

}
